@@ -1,4 +1,5 @@
-"""
+# -*- coding: utf-8 -*-
+u"""
 NetCDF reader/writer module.
 
 This module implements the Scientific.IO.NetCDF API to read and create
@@ -43,15 +44,20 @@ are automatically stored in the ``_attributes`` attribute by overloading
 ``obj.__dict__['key'] = value``, instead of simply ``obj.key = value``;
 otherwise the key would be inserted into userspace attributes.
 
+Unicode attribute values are allowed (although not required). This deals
+with the common use case of non-ASCII units, placenames, etc. Attribute
+values are encoded via UTF-8, as required by NetCDF and udunits2.
+
 To create a NetCDF file::
 
     >>> import time
     >>> f = netcdf_file('simple.nc', 'w')
     >>> f.history = 'Created for a test'
+    >>> f.location = u'北京'
     >>> f.createDimension('time', 10)
     >>> time = f.createVariable('time', 'i', ('time',))
     >>> time[:] = range(10)
-    >>> time.units = 'days since 2008-01-01'
+    >>> time.units = u'µs since 2008-01-01'
     >>> f.close()
 
 To read the NetCDF file we just created::
@@ -59,16 +65,22 @@ To read the NetCDF file we just created::
     >>> f = netcdf_file('simple.nc', 'r')
     >>> print f.history
     Created for a test
+    >>> print f.location
+    北京
     >>> time = f.variables['time']
     >>> print time.units
-    days since 2008-01-01
+    µs since 2008-01-01
     >>> print time.shape
     (10,)
     >>> print time[-1]
     9
     >>> f.close()
 
-TODO: properly implement ``_FillValue``.
+TODO:
+ * properly implement ``_FillValue``.
+ * implement Jeff Whitaker's patch for masked variables.
+ * fix character variables.
+ * implement PAGESIZE for Python 2.6?
 """
 
 __all__ = ['netcdf_file', 'netcdf_variable']
@@ -77,7 +89,7 @@ __all__ = ['netcdf_file', 'netcdf_variable']
 from operator import mul
 from mmap import mmap, ACCESS_READ
 
-from numpy import fromstring, ndarray, dtype, empty, array, asarray
+from numpy import fromstring, ndarray, dtype, empty, array, asarray, prod
 from numpy import little_endian as LITTLE_ENDIAN
 
 
@@ -127,8 +139,17 @@ class netcdf_file(object):
     attribute of the ``netcdf_file`` object.
 
     """
-    def __init__(self, filename, mode='r'):
-        self.filename = filename
+    def __init__(self, filename, mode='r', mmap=True, version=1):
+        if hasattr(filename, 'seek'):
+            self.fp = filename
+            self.filename = 'None'
+            self.use_mmap = False
+        else:
+            self.filename = filename
+            self.fp = open(self.filename, '%sb' % mode)
+            self.use_mmap = mmap
+
+        self.version_byte = version
 
         assert mode in 'rw', "Mode must be either 'r' or 'w'."
         self.mode = mode
@@ -140,8 +161,6 @@ class netcdf_file(object):
         self._recs = 0
         self._recsize = 0
 
-        self.fp = open(self.filename, '%sb' % mode)
-
         self._attributes = {}
 
         if mode is 'r':
@@ -188,9 +207,7 @@ class netcdf_file(object):
 
     def _write(self):
         self.fp.write('CDF')
-
-        self.__dict__['version_byte'] = 1
-        self.fp.write(array(1, '>b').tostring())
+        self.fp.write(array(self.version_byte, '>b').tostring())
 
         # Write headers and data.
         self._write_numrecs()
@@ -234,11 +251,11 @@ class netcdf_file(object):
             self.fp.write(NC_VARIABLE)
             self._pack_int(len(self.variables))
 
-            # Sort variables non-recs first, then recs.
-            variables = self.variables.items()
-            variables.sort(key=lambda (k, v): v._shape and not v.isrec)
-            variables.reverse()
-            variables = [k for (k, v) in variables]
+            # Sort variables non-recs first, then recs. We use a DSU 
+            # since some people use pupynere with Python 2.3.x.
+            deco = [ (v._shape and not v.isrec, k) for (k, v) in self.variables.items() ]
+            deco.sort()
+            variables = [ k for (unused, k) in deco ][::-1]
 
             # Set the metadata for all variables.
             for name in variables:
@@ -272,7 +289,10 @@ class netcdf_file(object):
             vsize = var.data.size * var.data.itemsize
             vsize += -vsize % 4
         else:  # record variable
-            vsize = var.data[0].size * var.data.itemsize
+            try:
+                vsize = var.data[0].size * var.data.itemsize
+            except IndexError:
+                vsize = 0
             rec_vars = len([var for var in self.variables.values()
                     if var.isrec])
             if rec_vars > 1:
@@ -321,10 +341,35 @@ class netcdf_file(object):
             self.fp.seek(pos0 + var._vsize)
 
     def _write_values(self, values):
-        values = asarray(values) 
-        values = values.astype(values.dtype.newbyteorder('>'))
+        if hasattr(values, 'dtype'):
+            nc_type = REVERSE[values.dtype.char]
+        else:
+            types = [
+                    (int, NC_INT),
+                    (long, NC_INT),
+                    (float, NC_FLOAT),
+                    (basestring, NC_CHAR),
+                    ]
+            try:
+                sample = values[0]
+            except (IndexError, TypeError):
+                sample = values
+            if isinstance(sample, unicode):
+                assert isinstance(values, unicode), type(values)
+                ## NetCDF requires that text be encoded via UTF-8
+                values = values.encode('utf-8')
+            for class_, nc_type in types:
+                if isinstance(sample, class_): break
+
+        typecode, size = TYPEMAP[nc_type]
+        if typecode is 'c':
+            dtype_ = '>c'
+        else:
+            dtype_ = '>%s' % typecode
+            if size > 1: dtype_ += str(size)
+
+        values = asarray(values, dtype=dtype_) 
 
-        nc_type = REVERSE[values.dtype.char]
         self.fp.write(nc_type)
 
         if values.dtype.char == 'S':
@@ -342,7 +387,8 @@ class netcdf_file(object):
 
     def _read(self):
         # Check magic bytes and version
-        assert self.fp.read(3) == 'CDF', "Error: %s is not a valid NetCDF 3 file" % self.filename
+        magic = self.fp.read(3)
+        assert magic == 'CDF', "Error: %s is not a valid NetCDF 3 file" % self.filename
         self.__dict__['version_byte'] = fromstring(self.fp.read(1), '>b')[0]
 
         # Read file headers and set data.
@@ -355,7 +401,8 @@ class netcdf_file(object):
         self.__dict__['_recs'] = self._unpack_int()
 
     def _read_dim_array(self):
-        assert self.fp.read(4) in [ZERO, NC_DIMENSION]
+        header = self.fp.read(4)
+        assert header in [ZERO, NC_DIMENSION]
         count = self._unpack_int()
 
         for dim in range(count):
@@ -369,7 +416,8 @@ class netcdf_file(object):
             self.__setattr__(k, v)
 
     def _read_att_array(self):
-        assert self.fp.read(4) in [ZERO, NC_ATTRIBUTE]
+        header = self.fp.read(4)
+        assert header in [ZERO, NC_ATTRIBUTE]
         count = self._unpack_int()
 
         attributes = {}
@@ -379,7 +427,8 @@ class netcdf_file(object):
         return attributes
 
     def _read_var_array(self):
-        assert self.fp.read(4) in [ZERO, NC_VARIABLE]
+        header = self.fp.read(4)
+        assert header in [ZERO, NC_VARIABLE]
 
         begin = 0
         dtypes = {'names': [], 'formats': []}
@@ -405,9 +454,17 @@ class netcdf_file(object):
                 # Data will be set later.
                 data = None
             else:
-                mm = mmap(self.fp.fileno(), begin_+vsize, access=ACCESS_READ)
-                data = ndarray.__new__(ndarray, shape, dtype=dtype_,
-                        buffer=mm, offset=begin_, order=0)
+                if self.use_mmap:
+                    mm = mmap(self.fp.fileno(), begin_+vsize, access=ACCESS_READ)
+                    data = ndarray.__new__(ndarray, shape, dtype=dtype_,
+                            buffer=mm, offset=begin_, order=0)
+                else:
+                    pos = self.fp.tell()
+                    self.fp.seek(begin_)
+                    #data = fromstring(self.fp.read(vsize), dtype=dtype_)
+                    data = fromstring(self.fp.read(size*prod(shape)), dtype=dtype_)
+                    data.shape = shape
+                    self.fp.seek(pos)
 
             # Add variable.
             self.variables[name] = netcdf_variable(
@@ -420,9 +477,16 @@ class netcdf_file(object):
                 dtypes['formats'] = dtypes['formats'][:1]
 
             # Build rec array.
-            mm = mmap(self.fp.fileno(), begin+self._recs*self._recsize, access=ACCESS_READ)
-            rec_array = ndarray.__new__(ndarray, (self._recs,), dtype=dtypes,
-                    buffer=mm, offset=begin, order=0)
+            if self.use_mmap:
+                mm = mmap(self.fp.fileno(), begin+self._recs*self._recsize, access=ACCESS_READ)
+                rec_array = ndarray.__new__(ndarray, (self._recs,), dtype=dtypes,
+                        buffer=mm, offset=begin, order=0)
+            else:
+                pos = self.fp.tell()
+                self.fp.seek(begin)
+                rec_array = fromstring(self.fp.read(self._recs*self._recsize), dtype=dtypes)
+                rec_array.shape = (self._recs,)
+                self.fp.seek(pos)
 
             for var in rec_vars:
                 self.variables[var].__dict__['data'] = rec_array[var]
@@ -470,7 +534,8 @@ class netcdf_file(object):
             values = fromstring(values, dtype='>%s%d' % (typecode, size))
             if values.shape == (1,): values = values[0]
         else:
-            values = values.rstrip('\x00') 
+            ## text values are encoded via UTF-8, per NetCDF standard
+            values = values.rstrip('\x00').decode('utf-8')
         return values
 
     def _pack_begin(self, begin):
@@ -544,19 +609,25 @@ class netcdf_variable(object):
             pass
         self.__dict__[attr] = value
 
-    @property
     def isrec(self):
         return self.data.shape and not self._shape[0]
+    isrec = property(isrec)
 
-    @property
     def shape(self):
         return self.data.shape
+    shape = property(shape)
 
     def getValue(self):
-        return self.data.item()
+        if self.data.shape:
+            return self.data[:]
+        else:
+            return self.data.item()
 
     def assignValue(self, value):
-        self.data.itemset(value)
+        if self.data.shape:
+            self.data[:] = value[:]
+        else:
+            self.data.itemset(value)
 
     def typecode(self):
         return self._typecode
